/*
Unnamed DNS Server (http://eder.us/projects/unnamed/
Copyright (C) 2012 David Eder

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define MAX_ENTRIES 200000

typedef struct
{
  uint32_t end;
  char label[16];
} country_t;

static int csv_get_field(char *buffer, char **begin, char **end, char **next)
{
  while(isspace((int)*buffer) != 0) buffer++;
  if(*buffer == '"' || *buffer == '\'')
  {
    *begin = buffer + 1;
    *end = buffer + 2;
    while(**end != *buffer && **end != '\0') (*end)++;
    *next = *end;
    if(**next != '\0')
    {
      (*next)++; /* skip quote */
      while(isspace((int)**next) != 0) (*next)++;
      if(**next == ',') (*next)++;
    }
  }
  else
  {
    *begin = *end = buffer;
    while(**end != ',' && **end != '\0') (*end)++;
    *next = *end;
    while(isspace((int)**end) != 0) (*end)--;
    if(**next != '\0')
    {
      (*next)++;  /* skip comma */
    }
  }
  while(isspace((int)**next) != 0) (*next)++;
  return 0;
}

static void set_country(country_t *country, uint32_t end, const char *label)
{
  int i;
  memset(country, 0, sizeof(country_t));
  country->end = end;
  for(i = 0; i < (int)sizeof(country->label) - 1 && label[i] != '\0'; i++)
  {
    country->label[i] = (char)tolower((int)label[i]);
  }
}

int main(int __attribute__((unused)) argc, char __attribute__((unused)) *argv[])
{
  country_t *country;
  char buffer[1024];
  char *begin = NULL, *end = NULL, *next = NULL;
  uint32_t begin_ip, end_ip;
  int ret = 0, i = 0;
  FILE *fp = fopen("GeoIPCountryWhois.csv", "r");
  if(fp == NULL) return -1;

  country = (country_t*)calloc(MAX_ENTRIES, sizeof(country_t));
  if(country == NULL)
  {
    (void)fclose(fp);
    return -1;
  }

  set_country(&country[i++], 0xffffff, "this"); /* 0.0.0.0/8 */

  while(i < MAX_ENTRIES && fgets(buffer, (int)sizeof(buffer), fp) != NULL)
  {
    (void)csv_get_field(buffer, &begin, &end, &next); /* text ip begin    */
    (void)csv_get_field(next, &begin, &end, &next);   /* text ip end      */
    (void)csv_get_field(next, &begin, &end, &next);   /* numeric ip begin */
    *end = '\0';
    begin_ip = (uint32_t)atoll(begin);
    (void)csv_get_field(next, &begin, &end, &next);   /* numeric ip end   */
    *end = '\0';
    end_ip = (uint32_t)atoll(begin);
    (void)csv_get_field(next, &begin, &end, &next);   /* label            */
    *end = '\0';

    while(i < MAX_ENTRIES && country[i-1].end + 1 != begin_ip)
    {
      switch(country[i-1].end + 1)
      {
        case 0x0a000000U: set_country(&country[i], (uint32_t)0x0affffffU, "private");   break; /* 10.0.0.0/8      */
        case 0x7f000000U: set_country(&country[i], (uint32_t)0x7fffffffU, "loopback");  break; /* 127.0.0.0/8     */
        case 0xa9fe0000U: set_country(&country[i], (uint32_t)0xa9feffffU, "linklocal"); break; /* 169.254.0.0/16  */
        case 0xac100000U: set_country(&country[i], (uint32_t)0xac1fffffU, "private");   break; /* 172.16.0.0/12   */
        case 0xc0000000U: set_country(&country[i], (uint32_t)0xc00000ffU, "ietf");      break; /* 192.0.0.0/24    */
        case 0xc0000200U: set_country(&country[i], (uint32_t)0xc00002ffU, "test");      break; /* 192.0.2.0/24    */
        case 0xc0586300U: set_country(&country[i], (uint32_t)0xc05863ffU, "6to4");      break; /* 192.88.99.0/24  */
        case 0xc0a80000U: set_country(&country[i], (uint32_t)0xc0a8ffffU, "private");   break; /* 192.168.0.0/16  */
        case 0xc0120000U: set_country(&country[i], (uint32_t)0xc013ffffU, "benchmark"); break; /* 192.18.0.0/15   */
        case 0xc6336400U: set_country(&country[i], (uint32_t)0xc63364ffU, "test");      break; /* 192.51.100.0/24 */
        case 0xcb007100U: set_country(&country[i], (uint32_t)0xcb0071ffU, "test");      break; /* 203.0.113.0/24  */
        case 0xe0000000U: set_country(&country[i], (uint32_t)0xefffffffU, "multicast"); break; /* 224.0.0.0/4     */
        default:          set_country(&country[i], begin_ip - 1, "unallocated");
      }
      i++;
    }
    if(i < MAX_ENTRIES)
    {
      set_country(&country[i++], end_ip, begin);
    }
    else
    {
      printf("i = %d, MAX_ENTRIES = %d\n", i, MAX_ENTRIES);
    }
  }
  if(i < MAX_ENTRIES)
  {
    set_country(&country[i++], 0xffffffff, "reserved");  /* 240.0.0.0/4 */
  }
  else
  {
    printf("i = %d, MAX_ENTRIES = %d\n", i, MAX_ENTRIES);
  }
  (void)fclose(fp);

  fp = fopen("country.db", "wb");
  if(fp != NULL)
  {
    if(fwrite(country, sizeof(country_t), (size_t)i, fp) != (size_t)i)
    {
      printf("Failed to write country.db\n");
      ret = -1;
    }
    (void)fclose(fp);
  }

  free(country);
  return ret;
}
