/*
Unnamed DNS Server (http://eder.us/projects/unnamed/
Copyright (C) 2012 David Eder

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifndef UNNAMED_H
#define UNNAMED_H

/* Possible things todo (in no particular order): */
/* spf is broken! */
/* unnamed_forward does not handle any correctly */
/* dnssec (requires EDNS) */
/* HINFO cpu=Tarissa / os=Skynet (rfc1035 section 3.3.2) */
/* WKS (rfc1035 section 3.4.2) */
/* in-addr.arpa subnets, partial subnets */
/* EDNS - RFC 2671 */
/* tcp forwarders */
/* tcp should not close until timeout, support multiple requests per session */
/* NSID - RFC 5001 (requires EDNS) */
/* NSEC3, NSEC3 whitelies*/
/* Thread pools */
/* cache to disk (unnamed_forward.c) */
/* SIGHUP to reload config, without flush of cache */
/* more documentation */
/* mdns / .local / Apple Bonjour / avahi / service discovery / zero conf */
/* rate limiting */
/* answer all to prevent spoof */
/* Resolve from expired cache on all else fails */
/* Recursor */
/* multiple simultaneous forward support */
/* dns crypt (opendns) */
/* zone transfer/notification IXFR=rfc1995 AXFR */
/* statistics */
/* lua module support for doing an internal dns lookup */
/* lua module support for reading/walking cache */
/* lua module support for arbitrary rr encoding */
/* lua example that reads /etc/resolv.conf */
/* lua example that reads /etc/hosts */
/* lua example that verifies ping before returning a/aaaa records */
/* lua example that resolves from database */
/* lua example that does dynamic dns hosting, TSIG (RFC 2845) */

#if !defined(WIN32) || defined(__CYGWIN__)
  #define USE_PTHREADS
#endif

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef USE_PTHREADS
  #include <pthread.h>
#endif

#ifdef WIN32
  #define FD_SETSIZE             1024
  #define SO_EXT                 "dll"
  #define WIN32_LEAN_AND_MEAN
  #include <direct.h>
  #include <windows.h>
  #include <winsock2.h>
  #include <ws2tcpip.h>
  #ifdef __MINGW32__
    #include <unistd.h>
    #include <dlfcn.h>
    #include <sys/time.h>
    #define DS                   "/"
  #else
    #include <direct.h>
    #define fileno(f)            _fileno(f)
    #define chdir(p)             _chdir(p)
    #define dlopen(f,m)          (void*)LoadLibrary((LPCSTR)f)
    #define dlsym(h,n)           GetProcAddress((HMODULE)h,n)
    #define dlclose(h)           (int)FreeLibrary(h)
    #define sleep(n)             Sleep(1000 * (n))
    #define snprintf             _snprintf
    #define strdup(s)            _strdup(s)
    #define chdir(d)             _chdir(d)
    #define chroot(p)            (((errno = ENOSYS) & 0) - 1)
    #define atoll(s)             _atoi64(s)
    #define __attribute__(a)
    typedef int ssize_t;
    #define DS                   "\\"
  #endif
  #define srandom(seed)          srand(seed)
  #define random()               rand()
  typedef int buf_len;
  typedef uint16_t in_port_t;
#else
  #include <unistd.h>
  #include <sys/socket.h>
  #include <sys/time.h>
  #include <netdb.h>
  #include <arpa/inet.h>
  #include <strings.h>
  #include <dlfcn.h>
  #define SO_EXT                 "so"
  #define DS                     "/"
  typedef int SOCKET;
  #define INVALID_SOCKET         -1
  #define closesocket(s)         close(s)
  typedef size_t buf_len;
#endif

#ifdef S_SPLINT_S
  #undef FD_ZERO
  #undef FD_SET
  #undef FD_ISSET
  #define FD_ZERO(fd)      memset(fd, 0, sizeof(*fd))
  #define FD_SET(fd,set) 
  #define FD_ISSET(fd,set) ((fd & 1) == 1)
#endif

#define MAX_QUESTIONS         10
#define MAX_RRS               16
#define MAX_LABELS           128

#define IPCHAR_SIZE           40
#define PACKET_LENGTH_SIZE     2

#define DEFAULT_TTL         7200

typedef struct
{
  uint8_t *value, small_value[32+sizeof(uint8_t*)-sizeof(uint32_t)];
  uint32_t length, offset, allocated;
} buffer_t;

#define buffer_init(b)  { (b)->allocated = (uint32_t)sizeof((b)->small_value); (b)->length = (b)->offset = 0; (b)->small_value[0] = 0; (b)->value = (b)->small_value; }

typedef struct
{
  union
  {
    struct sockaddr sockaddr;
    struct sockaddr_in sockaddr_in;
    struct sockaddr_in6 sockaddr_in6;
  } a;
} address_t;

typedef struct
{
  address_t begin, end;
} ip_range_t;

typedef struct _config_node
{
  struct _config_node *children;
  buffer_t name, value;
  int child_count, last, maxn;
} config_node_t;

typedef struct
{
  buffer_t name;
  int offset;
} name_t;

typedef struct
{
  uint16_t id;
  struct
  {
    #if __BYTE_ORDER == __BIG_ENDIAN
      unsigned int qr:1;
      unsigned int opcode:4;
      unsigned int aa:1;
      unsigned int tc:1;
      unsigned int rd:1;
      unsigned int ra:1;
      unsigned int z:1;
      unsigned int ad:1;
      unsigned int cd:1;
      unsigned int rcode:4;
    #elif __BYTE_ORDER == __LITTLE_ENDIAN
      unsigned int rd:1;
      unsigned int tc:1;
      unsigned int aa:1;
      unsigned int opcode:4;
      unsigned int qr:1;
      unsigned int rcode:4;
      unsigned int cd:1;
      unsigned int ad:1;
      unsigned int z:1;
      unsigned int ra:1;
    #endif
  } flags;
  uint16_t qdcount;
  uint16_t ancount;
  uint16_t nscount;
  uint16_t arcount;
} header_t;

typedef struct
{
  name_t qname;
  uint16_t qtype, qclass;
} question_t;

typedef struct _rr
{
  name_t name;
  struct _rr *next;
  uint32_t expires; /* absolute time, time + ttl */
  uint16_t rtype, rclass;
  union
  {
    address_t address;
    name_t cname;
    struct
    {
      name_t exchange;
      uint16_t preference;
    } mx;
    name_t ns;
    name_t ptr;
    struct
    {
      name_t mname, rname;
      uint32_t serial, refresh, retry, expire, minimum;
    } soa;
    buffer_t spf;
    buffer_t txt;
    struct
    {
      uint16_t priority;
      uint16_t weight;
      uint16_t port;
      name_t target;
    } srv;
  } p;
  uint8_t status;
} rr_t;

typedef enum
{
  RR_AN, RR_NS, RR_AR
} rr_section_t;

typedef enum
{
  ANSWER_SATISFIED,
  ANSWER_UNSATISFIED,
  ANSWER_BLOCKED,
  ANSWER_NOTFOUND,
  ANSWER_NAMEERROR
} answer_t;

typedef struct treap_node
{
  char *key;
  struct treap_node *left, *right;
  int priority;
} treap_t;

typedef struct
{
  char **patterns;
  treap_t *contains, *begins, *ends, *whole;
  int pattern_count;
} pattern_list_t;

struct _state_t;
typedef struct _state_t state_t;

typedef struct
{
  buffer_t packet;
  state_t *state;
  SOCKET socket;
  address_t from_addr, to_addr;
  header_t header;
  question_t *questions[MAX_QUESTIONS];
  rr_t *answers[MAX_RRS];
  rr_t *name_servers[MAX_RRS];
  rr_t *additional_records[MAX_RRS];
  double start;
  uint16_t max_udp_size;
} client_t;

typedef struct
{
  const char *name;
  char *so;
  int (*init)(int, state_t*);
  answer_t (*answer)(int, client_t *, rr_section_t, const buffer_t *, uint16_t, uint16_t, int);
  void (*collect_garbage)(int, state_t*);
  void (*term)(int, state_t*);
} module_t;

struct _state_t
{
  config_node_t config;

  module_t *modules;
  void **module_data;
  void **module_handle;
  int module_count;

  uint32_t min_ttl;
  FILE *lprintf_fp;
  pattern_list_t white_list, black_list;
  ip_range_t *blacklist_range;
  ip_range_t *allow_forwards;
  int blacklist_range_count;
  int allow_forwards_count;
  int trace;
  int random_fd;

  void (*buffer_free)(buffer_t *b);
  int (*buffer_resize)(buffer_t *b, uint32_t size);
  int (*buffer_copy)(buffer_t *dst, const buffer_t *src);
  int (*buffer_put_raw)(buffer_t *b, const uint8_t *src, uint32_t length);
  int (*buffer_set)(buffer_t *b, const char *fmt, ...) __attribute__ ((format (printf, 2, 3)));
  int (*buffer_cmp)(const buffer_t *a, const buffer_t *b);

  const char* (*addr2char)(char ip[IPCHAR_SIZE], const address_t *a);
  int (*char2addr)(const char *ip, in_port_t port, /*@out@*/ address_t *addr);
  int (*addr_cmp)(const address_t *a, const address_t *b);
  int (*is_in_range)(const address_t *address, const ip_range_t *range);
  int (*addr6_prefix)(const buffer_t *name, /*@out@*/ address_t *a);
  int (*addr_prefix)(const buffer_t *name, /*@out@*/ address_t *a);
  int (*is_subdomain)(const char *domain, const char *check);
  int (*ptr_name)(buffer_t *name, const address_t *a);

  const char* (*config_get_const)(state_t *state, const char *name);
  char* (*config_get)(state_t *state, const char *name);
  char* (*config_get_lower)(state_t *state, const char *name);
  int (*config_get_int)(state_t *state, const char *name, int default_value);
  double (*config_get_double)(state_t *state, const char *name, double default_value);
  int (*config_get_ip)(state_t *state, const char *name,  address_t *addr);
  int (*config_get_ip_range)(state_t *state, const char *name, ip_range_t *range);
  void (*config_free_node)(state_t *state, const char *name);
  int (*config_maxn)(state_t *state, const char *name);

  int (*is_ip_black)(state_t *state, const address_t *a);
  int (*is_name_white)(state_t *state, const buffer_t *name);
  int (*is_name_black)(state_t *state, const buffer_t *name);

  question_t* (*question_create)(const buffer_t *name, uint16_t type, uint16_t dclass);
  void (*free_rr)(rr_t **rr);
  int (*add_rr)(client_t *client, rr_section_t rr_section, rr_t **rr);
  rr_t* (*rr_create)(const buffer_t *name, uint16_t type, uint16_t dclass, uint32_t expires, uint8_t status);
  int (*parse_packet)(client_t *client);
  int (*write_packet)(client_t *client);
  int (*init_client)(state_t *state, client_t *client, uint16_t max_udp_size);
  void* (*free_client)(client_t *client, int dynamic);
  double (*microtime)(void);
  uint32_t (*current_serial)(void);
  __attribute__((const)) const char* (*type_name)(uint16_t type);
  __attribute__((const)) const char* (*class_name)(uint16_t dclass);
  __attribute__((const)) const char* (*answer_type_name)(answer_t a);
  __attribute__((const)) const char* (*rcode_name)(uint8_t rcode);
  answer_t (*answer_question)(client_t *client, rr_section_t rr_section, const buffer_t *name, uint16_t type, uint16_t dclass, int lazy);
  int (*match)(const char *pattern, const char *string);

  uint16_t (*random_id)(state_t *state);

  int (*lprintf)(state_t *state, const char *file, const char *function, int line, const char *fmt, ...) __attribute__ ((format (printf, 5, 6)));
};

#define ERROR_IF(trace)    if((trace) >= TRACE_LEVEL_ERROR)
#define WARNING_IF(trace)  if((trace) >= TRACE_LEVEL_WARNING)
#define INFO_IF(trace)     if((trace) >= TRACE_LEVEL_INFO)
#define DEBUG_IF(trace)    if((trace) >= TRACE_LEVEL_DEBUG)

#define TRACE   state, __FILE__, __FUNCTION__, __LINE__

#define TRACE_LEVEL_ERROR      0
#define TRACE_LEVEL_WARNING    1
#define TRACE_LEVEL_INFO       2
#define TRACE_LEVEL_DEBUG      3

/* #define ERRNO_STRING  ((errno >= 0 && errno < sys_nerr) ? sys_errlist[errno] : "unknown error") */
#define ERRNO_STRING  strerror(errno)

#define CLASS_IN               1    /* RFC 1035 */
#define CLASS_CH               3    /* RFC 1035 */
#define CLASS_HESIOD           4
#define CLASS_NONE           254    /* RFC 2136 */
#define CLASS_ANY            255    /* RFC 1035 */

#define TYPE_A                 1    /* RFC 1035 */
#define TYPE_NS                2    /* RFC 1035 */
#define TYPE_MD                3    /* RFC 1035 */
#define TYPE_MF                4    /* RFC 1035 */
#define TYPE_CNAME             5    /* RFC 1035 */
#define TYPE_SOA               6    /* RFC 1035 */
#define TYPE_MB                7    /* RFC 1035 */
#define TYPE_MG                8    /* RFC 1035 */
#define TYPE_MR                9    /* RFC 1035 */
#define TYPE_NULL             10    /* RFC 1035 */
#define TYPE_WKS              11    /* RFC 1035 */
#define TYPE_PTR              12    /* RFC 1035 */
#define TYPE_HINFO            13    /* RFC 1035 */
#define TYPE_MINFO            14    /* RFC 1035 */
#define TYPE_MX               15    /* RFC 1035 */
#define TYPE_TXT              16    /* RFC 1035 */
#define TYPE_RP               17    /* RFC 1183 */
#define TYPE_AFSDB            18    /* RFC 1183 */
#define TYPE_X25              19    /* RFC 1183 */
#define TYPE_ISDN             20    /* RFC 1183 */
#define TYPE_RT               21    /* RFC 1183 */
#define TYPE_NSAP             22    /* RFC 1706 */
#define TYPE_NSAPPTR          23    /* RFC 1348 */
#define TYPE_SIG              24    /* RFC 4034, RFC 3755, RFC 2535 */
#define TYPE_KEY              25    /* RFC 4034, RFC 3755, RFC 2535 */
#define TYPE_PX               26    /* RFC 2163 */
#define TYPE_GPOS             27    /* RFC 1712 */
#define TYPE_AAAA             28    /* RFC 3596 */
#define TYPE_LOC              29    /* RFC 1876 */
#define TYPE_NXT              30    /* RFC 3755, RFC 2535 */
#define TYPE_EID              31
#define TYPE_NIMLOC           32
#define TYPE_SRV              33    /* RFC 2782 */
#define TYPE_ATMA             34
#define TYPE_NAPTR            35    /* RFC 3403 */
#define TYPE_KX               36    /* RFC 2230 */
#define TYPE_CERT             37    /* RFC 4398 */
#define TYPE_A6               38    /* RFC 3226, RFC 2874 */
#define TYPE_DNAME            39    /* RFC 2672 */
#define TYPE_SINK             40
#define TYPE_OPT              41    /* RFC 2671 */
#define TYPE_APL              42    /* RFC 3123 */
#define TYPE_DS               43    /* RFC 4034 */
#define TYPE_SSHFP            44    /* RFC 4255 */
#define TYPE_IPSECKEY         45    /* RFC 4025 */
#define TYPE_RRSIG            46    /* RFC 4034 */
#define TYPE_NSEC             47    /* RFC 4034 */
#define TYPE_DNSKEY           48    /* RFC 4034 */
#define TYPE_DHCID            49    /* RFC 4701 */
#define TYPE_NSEC3            50    /* RFC 5155 */
#define TYPE_NSEC3PARAM       51    /* RFC 5155 */
#define TYPE_HIP              55    /* RFC 5205 */
#define TYPE_NINFO            56
#define TYPE_RKEY             57
#define TYPE_TALINK           58
#define TYPE_CDS              59
#define TYPE_SPF              99    /* RFC 4408 */
#define TYPE_UINFO           100
#define TYPE_UID             101
#define TYPE_GID             102
#define TYPE_UNSPEC          103
#define TYPE_TKEY            249    /* RFC 2930 */
#define TYPE_TSIG            250    /* RFC 2845 */
#define TYPE_IXFR            251    /* RFC 1995 */
#define TYPE_AXFR            252    /* RFC 1035 */
#define TYPE_MAILB           253
#define TYPE_MAILA           254
#define TYPE_ANY             255    /* RFC 1035 */
#define TYPE_URI             256
#define TYPE_CAA             257
#define TYPE_TA            32768
#define TYPE_DLV           32769    /* RFC 4431 */

#define OPCODE_QUERY           0
#define OPCODE_IQUERY          1
#define OPCODE_STATUS          2

#define RCODE_NO_ERROR         0
#define RCODE_FORMAT_ERROR     1
#define RCODE_SERVER_FAILURE   2
#define RCODE_NAME_ERROR       3
#define RCODE_NOT_IMPLEMENTED  4
#define RCODE_REFUSED          5

#define RCODE_NOT_FOUND        0xff

#endif
