/*
Unnamed DNS Server (http://eder.us/projects/unnamed/
Copyright (C) 2012 David Eder

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "unnamed.h"

typedef struct
{
  uint32_t ttl;
  int trace;
} module_data_t;

static int init(int module_index, state_t *state)
{
  module_data_t *module = (module_data_t*)calloc(1, sizeof(*module));
  if(module == NULL) return -1;
  state->module_data[module_index] = module;
  module->trace = state->config_get_int(state, "ipv6_literal.trace", state->trace);
  module->ttl = (uint32_t)state->config_get_int(state, "ipv6_literal.ttl", DEFAULT_TTL);
  state->config_free_node(state, "ipv6_literal");
  return 0;
}

static void term(int module_index, state_t *state)
{
  module_data_t *module = (module_data_t*)state->module_data[module_index];
  free(module);
  state->module_data[module_index] = NULL;
}

static answer_t answer(int module_index, client_t *client, rr_section_t rr_section, const buffer_t *name, uint16_t type, uint16_t dclass, int lazy)
{
  buffer_t address;
  address_t a;
  int i;
  state_t *state = client->state;
  module_data_t *module = (module_data_t*)state->module_data[module_index];

  if(lazy != 0) return ANSWER_UNSATISFIED;
  if(dclass != CLASS_IN && dclass != CLASS_ANY) return ANSWER_UNSATISFIED;
  if(type != TYPE_AAAA && type != TYPE_ANY) return ANSWER_UNSATISFIED;
  if(client->header.flags.ra == 0) return ANSWER_UNSATISFIED;
  if(state->is_subdomain("ipv6-literal.net", (char*)name->value) == 0) return ANSWER_UNSATISFIED;

  buffer_init(&address);

  if(state->buffer_copy(&address, name) != 0) return ANSWER_UNSATISFIED;
  for(i = 0; i < (int)address.length; i++)
  {
    if(address.value[i] == (uint8_t)'-') address.value[i] = (uint8_t)':';
    else if(address.value[i] == (uint8_t)'.' || address.value[i] == (uint8_t)'s') break;
  }
  address.value[i] = (uint8_t)'\0';
  address.length = (uint32_t)i;

  if(state->char2addr((char*)address.value, 0, &a) == 0)
  {
    rr_t *rr;
    if((rr = state->rr_create(name, TYPE_AAAA, CLASS_IN, module->ttl, RCODE_NO_ERROR)) != NULL)
    {
      memcpy(&rr->p.address, &a, sizeof(rr->p.address));
      if(state->add_rr(client, rr_section, &rr) == 0)
      {
        char addr_temp[IPCHAR_SIZE];
        addr_temp[0] = '\0';
        state->free_rr(&rr);
        INFO_IF(module->trace) (void)state->lprintf(TRACE, "%s %s %s => %s", (char*)name->value, state->class_name(dclass), state->type_name(type), state->addr2char(addr_temp, &a));
        return ANSWER_SATISFIED;
      }
      state->free_rr(&rr);
    }
  }
  return ANSWER_UNSATISFIED;
}

module_t unnamed_module = { "ipv6_literal", NULL, init, answer, NULL, term };
