/*
Unnamed DNS Server (http://eder.us/projects/unnamed/
Copyright (C) 2012 David Eder

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdio.h>
#include <fcntl.h>
#include "unnamed_internal.h"


/* PHASE 1, make requests */
/* no response is considered an error! */
/* tcp invalid length */
/* encode labels wrong */
/* include an, ns, ar in request */
/* play with header bits */
/* random bits changed any where */
/* udp too long */
/* a, aaaa, mx, cname, ptr, spf, txt */
/* unknown class */
/* unknown type */

/* PHASE 2, become a fowarder */

/* same games as above */
/* rr wrong length */
/* rr length longer than packet */

/* PHASE 3, thread stress */

static void generate_test(const char *domain, uint16_t type, uint16_t dclass, int bits_to_fiddle)
{
  int n;
  SOCKET sock;
  uint16_t id = 0;
  struct timeval tv;
  fd_set rfds;
  client_t client;
  address_t addr, reply;
  socklen_t reply_len = (socklen_t)sizeof(reply);
  buffer_t name;
  state_t _state, *state;
  state = &_state;
  memset(state, 0, sizeof(*state));
  state->trace = TRACE_LEVEL_INFO;
  state->lprintf = lprintf;
  state->random_fd = -1;

  memset(&tv, 0, sizeof(tv));

  (void)char2addr("127.0.0.1", 53, &addr);
  buffer_init(&name);

  (void)buffer_set(&name, "%s", domain);

  if((sock = socket(AF_INET, SOCK_DGRAM, 0)) == INVALID_SOCKET)
  {
    ERROR_IF(state->trace) (void)lprintf(TRACE, "unnamed_socket failed.");
    return;
  }

  memset(&client, 0, sizeof(client));
  if(init_client(state, &client, 512) != 0)
  {
    ERROR_IF(state->trace) (void)lprintf(TRACE, "init_client failed.");
    (void)closesocket(sock);
    (void)state->free_client(&client, 0);
    return;
  }

  client.header.id = random_id(state);
  client.header.flags.rd = 1;
  client.header.qdcount = 0;
  client.header.ancount = 0;
  client.header.nscount = 0;
  client.header.arcount = 0;

  if((client.questions[client.header.qdcount++] = question_create(&name, type, dclass)) == NULL)
  {
    (void)closesocket(sock);
    (void)free_client(&client, 0);
    return;
  }

  if(write_packet(&client) != 0)
  {
    (void)closesocket(sock);
    (void)free_client(&client, 0);
    return;
  }

  while(bits_to_fiddle-- > 0)
  {
    client.packet.value[(uint32_t)random() % client.packet.length] ^= (uint8_t)(1 << (unsigned int)(random() % 8));
  }

  if((int)sendto(sock, (char*)(client.packet.value + PACKET_LENGTH_SIZE), (buf_len)client.packet.length - PACKET_LENGTH_SIZE, 0, &addr.a.sockaddr, (socklen_t)sizeof(addr)) != (int)client.packet.length - PACKET_LENGTH_SIZE)
  {
    ERROR_IF(state->trace) (void)lprintf(TRACE, "sendto failed: %s", ERRNO_STRING);
    (void)closesocket(sock);
    (void)free_client(&client, 0);
    return;
  }

  FD_ZERO(&rfds);
  FD_SET((unsigned int)sock, &rfds);

  (void)buffer_resize(&client.packet, 512);

  tv.tv_sec = 1;
  tv.tv_usec = 0;
  n = select((int)sock+1, &rfds, NULL, NULL, &tv);
  if(n == -1)
  {
    ERROR_IF(state->trace) (void)lprintf(TRACE, "select failed: %s", ERRNO_STRING);
    (void)closesocket(sock);
    (void)free_client(&client, 0);
    return;
  }
  if(n == 0) /* timeout */
  {
printf("timeout\n");
    (void)closesocket(sock);
    (void)free_client(&client, 0);
    return;
  }

  memset(&reply, 0, sizeof(reply));
  if((n = (int)recvfrom(sock, (char*)client.packet.value + PACKET_LENGTH_SIZE, (buf_len)(client.packet.allocated - PACKET_LENGTH_SIZE), 0, (struct sockaddr*)&reply, &reply_len)) == -1)
  {
    ERROR_IF(state->trace) (void)lprintf(TRACE, "recvfrom failed: %s", ERRNO_STRING);
    (void)closesocket(sock);
    (void)free_client(&client, 0);
    return;
  }
  (void)closesocket(sock);
  client.packet.value[0] = (uint8_t)(((unsigned int)n >> 8) & 0xff);
  client.packet.value[1] = (uint8_t)(n & 0xff);
  client.packet.length = (uint32_t)n + 2;

  if(parse_packet(&client) != 0)
  {
    int i;
    WARNING_IF(state->trace) (void)lprintf(TRACE, "Unable to parse response");
    printf("packet:"); for(i = 0; i < (int)client.packet.length; i++) printf(" %02x", (unsigned int)client.packet.value[i]); printf("\n");
    (void)free_client(&client, 0);
    return;
  }

  if(id != client.header.id)
  {
    WARNING_IF(state->trace) (void)lprintf(TRACE, "ID mismatch %u != %u", (unsigned int)client.header.id, (unsigned int)id);
    (void)free_client(&client, 0);
    return;
  }

  (void)free_client(&client, 0);
}


int main(int __attribute__((unused)) argc, char __attribute__((unused)) *argv[])
{
  int i;
  struct timeval tv;

  #ifdef _MSC_VER
    FILETIME ft;
    unsigned __int64 usec;
    GetSystemTimeAsFileTime(&ft);
    usec = ((((__int64)ft.dwHighDateTime) << 32) | (__int64)ft.dwLowDateTime) / 10Ui64 - 11644473600000000Ui64;
    tv.tv_sec = (long)(usec / 1000000UL);
    tv.tv_usec = (long)(usec % 1000000UL);
  #else
    memset(&tv, 0, sizeof(tv));
    (void)gettimeofday(&tv, NULL);
  #endif
  srandom((unsigned int)tv.tv_usec);

  for(i = 0; i < 1000; i++)
  {
    printf("%d\n", i);
    generate_test("eder.us", TYPE_A, CLASS_IN, (int)random() % 8);
    generate_test("google.com", TYPE_A, CLASS_IN, (int)random() % 8);
  }
  return 0;
}
